@@ -6,7 +6,7 @@ import inspect
 import json
 import logging
 import uuid
-from typing import Union, Optional
+from typing import Union, Optional, Tuple
 
 from gremlinapi.exceptions import (
     GremlinCommandTargetError,
@@ -18,6 +18,7 @@ from gremlinapi.attack_helpers import (
     GremlinAttackCommandHelper,
     GremlinAttackTargetHelper,
 )
+from gremlinapi.util import deprecated
 from gremlinapi.clients import GremlinAPIClients as clients
 from gremlinapi.containers import GremlinAPIContainers as containers
 from gremlinapi.providers import GremlinAPIProviders as providers
@@ -35,14 +36,33 @@ class GremlinScenarioNode(object):
     ):
         self._edges: dict = dict()
         self._id: str = str()
-        self._previous: "GremlinScenarioNode" = None  # type: ignore
         self._name: str = str()
-        self._next: "GremlinScenarioNode" = None  # type: ignore
         self._node_type: str = str()
         self.id: str = str(uuid.uuid4())
+        self._index: str = str()
+        self.index = "0"
+        self._next: str = str()
+        self.next = "0"
         self.name: str = kwargs.get("name", None)  # type: ignore
 
     def add_edge(self, _node: "GremlinScenarioNode", _weight: str = None) -> None:
+        """
+        Adds an edge from this node to the value of _node.
+
+        This method contains the data structure for an edge.
+
+        Parameters
+        ----------
+        _node : GremlinScenarioNode
+            The node for which an edge needs to be added
+        _weight : int optional
+            The optional weight of the edge
+
+        Raises
+        ------
+        GremlinParameterError
+            If the _node is not of the type GremlinScenarioNode
+        """
         if not issubclass(type(_node), GremlinScenarioNode):
             error_msg: str = (
                 f"add_edge expects a GremlinScenarioNode, received {type(_node)}"
@@ -62,11 +82,39 @@ class GremlinScenarioNode(object):
     @id.setter
     def id(self, _id: str = None) -> None:
         if not isinstance(_id, str):
-            error_msg: str = f"id expects a string {type(str)}, received {type(str)}"
+            error_msg: str = f"id expects a string {type(str)}, received {type(_id)}"
             log.error(error_msg)
             raise GremlinParameterError(error_msg)
         self._id = _id
 
+    @property
+    def index(self) -> str:
+        return self._index
+
+    @index.setter
+    def index(self, _index: str = "0") -> None:
+        if not isinstance(_index, str):
+            error_msg: str = (
+                f"index expects an string {type(str)}, received {type(_index)}"
+            )
+            log.error(error_msg)
+            raise GremlinParameterError(error_msg)
+        self._index = _index
+
+    @property
+    def next(self) -> str:
+        return self._next
+
+    @next.setter
+    def next(self, _next: str = "0") -> None:
+        if not isinstance(_next, str):
+            error_msg: str = (
+                f"next expects an string {type(str)}, received {type(_next)}"
+            )
+            log.error(error_msg)
+            raise GremlinParameterError(error_msg)
+        self._next = _next
+
     @property
     def name(self) -> str:
         return self._name
@@ -81,20 +129,6 @@ class GremlinScenarioNode(object):
             raise GremlinParameterError(error_msg)
         self._name = _name
 
-    @property
-    def next(self) -> "GremlinScenarioNode":
-        return self._next
-
-    @next.setter
-    def next(self, _node: "GremlinScenarioNode") -> None:
-        if not issubclass(type(_node), GremlinScenarioNode):
-            error_msg: str = (
-                f"next expects a GremlinScenarioNode, received {type(_node)}"
-            )
-            log.error(error_msg)
-            raise GremlinParameterError(error_msg)
-        self._next = _node
-
     @property
     def node_type(self) -> str:
         return self._node_type
@@ -109,20 +143,6 @@ class GremlinScenarioNode(object):
             raise GremlinParameterError(error_msg)
         self._node_type = _node_type
 
-    @property
-    def previous(self) -> "GremlinScenarioNode":
-        return self._previous
-
-    @previous.setter
-    def previous(self, _node: "GremlinScenarioNode") -> None:
-        if not issubclass(type(_node), GremlinScenarioNode):
-            error_msg: str = (
-                f"previous expects a GremlinScenarioNode, received {type(_node)}"
-            )
-            log.error(error_msg)
-            raise GremlinParameterError(error_msg)
-        self._previous = _node
-
     @property
     def uuid(self) -> str:
         if not self.name:
@@ -132,10 +152,12 @@ class GremlinScenarioNode(object):
         return f"{self.name}-{self.id}"
 
     def repr_model(self) -> dict:
-        if not self.next:
-            model = {"type": self.node_type, "id": self.uuid, "next": None}
-        else:
-            model = {"type": self.node_type, "id": self.uuid, "next": self.next.uuid}
+        model = {
+            "type": self.node_type,
+            "guid": self.uuid,
+            "id": self.index,
+            "next": self.next,
+        }
         return model
 
     def __repr__(self) -> str:
@@ -157,12 +179,125 @@ class GremlinScenarioGraphHelper(object):
         self.hypothesis: str = kwargs.get("hypothesis", None)  # type: ignore
         self.name: str = kwargs.get("name", None)  # type: ignore
 
-    def add_node(self, _node: GremlinScenarioNode) -> None:
-        if not issubclass(type(_node), GremlinScenarioNode):
-            error_msg: str = f"add_node expects GremlinScenarioNode (or None), received {type(_node)}"
+    def add_node(self, node: GremlinScenarioNode, _default_edge: bool = False) -> None:
+        """
+        Adds the node to the graph as a loose leaf. If it is the first node, it becomes the HEAD
+        node.
+
+        Parameters
+        ----------
+        node : GremlinScenarioNode
+            The node to add to the graph
+        _default_edge bool optional
+            Adds an edge from node to the last-added node in the graph
+
+        Raises
+        ------
+        GremlinParameterError
+            If the node is not of the type GremlinScenarioNode
+        """
+        if not issubclass(type(node), GremlinScenarioNode):
+            error_msg: str = (
+                f"add_node expects GremlinScenarioNode (or None), received {type(node)}"
+            )
             log.error(error_msg)
             raise GremlinParameterError(error_msg)
-        self._nodes.append(_node)
+        if not self._nodes.head:
+            _default_edge = False
+        if _default_edge:
+            tail_node = self._nodes._nodes[-1]
+        self._nodes.append(node)
+        if _default_edge:
+            self.add_edge(node, tail_node)
+
+    def remove_node(self, node: GremlinScenarioNode) -> None:
+        """
+        Removes node from graph.
+
+        If the node to remove is the head node, Raises an error.
+
+        Parameters
+        ----------
+        node : GremlinScenarioNode
+            Node to remove from the graph
+
+        Raises
+        ------
+        GremlinParameterError
+            If the _node to remove is the current head node
+        """
+        self._nodes.remove(node)
+
+    def get_last_node(self) -> GremlinScenarioNode:
+        return self._nodes._nodes[-1]
+
+    def add_edge(
+        self,
+        dst_node: GremlinScenarioNode,
+        _src_node: GremlinScenarioNode = None,
+        _weight: int = None,
+    ) -> None:
+        """
+        Helper function to add edges to source and destination nodes, with optional weight.
+
+        If _src_node is not defined, the node with the highest id (the last added)
+        is used
+
+        Parameters
+        ----------
+        dst_node : GremlinScenarioNode
+            The node for which an edge needs to be added.
+        _src_node : GremlinScenatioNode optional
+            The optional source node if a custom edge is required.
+        _weight : int optional
+            The optional weight of the edge
+
+        Raises
+        ------
+        GremlinParameterError
+            If the _src_node or dst_node are not of the type GremlinScenarioNode
+        """
+        if not issubclass(type(dst_node), GremlinScenarioNode):
+            error_msg: str = f"add_edge expects GremlinScenarioNode (or None), received {type(dst_node)}"
+            log.error(error_msg)
+            raise GremlinParameterError(error_msg)
+        if not _src_node:
+            _src_node = self.get_last_node()
+        self._nodes.add_edge(_src_node, dst_node, _weight)
+
+    def remove_edge(
+        self,
+        src_node: GremlinScenarioNode,
+        _edge_node: GremlinScenarioNode = None,
+    ) -> None:
+        """
+        Helper function to remove edges from nodes
+
+        If _edge_node is not defined, all edges from src_node will be removed from the graph
+
+        Parameters
+        ----------
+        src_node : GremlinScenarioNode
+            One side of the edge to be removed
+        _edge_node : GremlinScenatioNode optional
+            The optional other side of the edge to be removed
+
+        Raises
+        ------
+        GremlinParameterError
+            If src_node is not of the type GremlinScenarioNode
+        """
+        if not issubclass(type(src_node), GremlinScenarioNode):
+            error_msg: str = f"remove_edge expects GremlinScenarioNode (or None), received {type(src_node)}"
+            log.error(error_msg)
+            raise GremlinParameterError(error_msg)
+
+        if not _edge_node:
+            src_edges = src_node._edges.copy()
+            for node_id in src_edges:
+                self._nodes.remove_edge(src_node, src_node._edges[node_id]["node"])
+        else:
+            self._nodes.remove_edge(src_node, _edge_node)
 
     @property
     def description(self) -> str:
@@ -212,10 +347,11 @@ class GremlinScenarioGraphHelper(object):
         }
         if self._nodes.head is not None:
             model["graph"] = {
-                "start_id": self._nodes.head.uuid,
-                "nodes": {
-                    node.uuid: data for node, data in self._nodes.nodes_data_linear()
-                },
+                "start_id": "0",
+                "nodes": self._nodes.get_nodes_linear()
+                # "nodes": {
+                #     node.uuid: data for node, data in self._nodes.nodes_data_linear()
+                # }
             }
         return model
 
@@ -224,7 +360,26 @@ class GremlinScenarioGraphHelper(object):
         return json.dumps(model)
 
 
-class GremlinScenarioAttackNode(GremlinScenarioNode):
+class GremlinScenarioSerialNode(GremlinScenarioNode):
+    def __init(
+        self,
+        *args: tuple,
+        **kwargs: dict,
+    ):
+        super().__init__(*args, **kwargs)
+
+
+class GremlinScenarioParallelNode(GremlinScenarioNode):
+    def __init(
+        self,
+        *args: tuple,
+        **kwargs: dict,
+    ):
+        super().__init__(*args, **kwargs)
+        raise NotImplementedError("Parallel Scenario Nodes NOT IMPLEMENTED")
+
+
+class GremlinScenarioAttackNode(GremlinScenarioSerialNode):
     def __init__(
         self,
         *args: tuple,
@@ -249,7 +404,7 @@ class GremlinScenarioAttackNode(GremlinScenarioNode):
         self._attack_type = _attack_type
 
 
-class GremlinScenarioILFINode(GremlinScenarioAttackNode):
+class GremlinScenarioILFINode(GremlinScenarioSerialNode):
     def __init__(self, *args: tuple, **kwargs: dict):
         if not kwargs.get("name", None) and kwargs.get("command", None):  # type: ignore
             kwargs["name"] = (kwargs.get("command")).shortType  # type: ignore
@@ -291,7 +446,7 @@ class GremlinScenarioILFINode(GremlinScenarioAttackNode):
         self._target = _target
 
 
-class GremlinScenarioALFINode(GremlinScenarioAttackNode):
+class GremlinScenarioALFINode(GremlinScenarioSerialNode):
     def __init__(
         self,
         *args: tuple,
@@ -302,7 +457,7 @@ class GremlinScenarioALFINode(GremlinScenarioAttackNode):
         raise NotImplementedError("ALFI Scenarios NOT IMPLEMENTED")
 
 
-class GremlinScenarioDelayNode(GremlinScenarioNode):
+class GremlinScenarioDelayNode(GremlinScenarioSerialNode):
     def __init__(self, *args: tuple, **kwargs: dict):
         if not kwargs.get("name", None):
             kwargs["name"] = "Delay"  # type: ignore
@@ -331,7 +486,7 @@ class GremlinScenarioDelayNode(GremlinScenarioNode):
         self._delay = _duration
 
 
-class GremlinScenarioStatusCheckNode(GremlinScenarioNode):
+class GremlinScenarioStatusCheckNode(GremlinScenarioSerialNode):
     def __init__(
         self,
         *args: tuple,
@@ -450,97 +605,225 @@ class GremlinScenarioStatusCheckNode(GremlinScenarioNode):
 class _GremlinNodeGraph(object):
     def __init__(self):
         self._head: GremlinScenarioNode = None  # type: ignore
+        self._nodes: list = []
 
     def add_edge(
         self,
-        src_node: GremlinScenarioNode,
-        dest_node: GremlinScenarioNode,
+        node_left: GremlinScenarioNode,
+        node_right: GremlinScenarioNode,
         _weight=None,
     ) -> None:
-        self._validate_type(src_node)
-        self._validate_type(dest_node)
-        src_node.add_edge(dest_node, _weight)
-        dest_node.add_edge(src_node, _weight)
+        """
+        Adds edges to left and right nodes, with optional weight
+
+        Parameters
+        ----------
+        node_left : GremlinScenarioNode
+            One side of the new edge
+        node_right : GremlinScenarioNode
+            One side of the new edge
+        _weight : int optional
+            The optional weight of the edge
+
+        Raises
+        ------
+        GremlinParameterError
+            If the node_left or node_right are not of the type GremlinScenarioNode
+        """
+        self._validate_type(node_left)
+        self._validate_type(node_right)
+        node_left.add_edge(node_right, _weight)
+        node_right.add_edge(node_left, _weight)
+
+    def remove_edge(
+        self,
+        node_left: GremlinScenarioNode,
+        node_right: GremlinScenarioNode,
+    ) -> None:
+        """
+        Removes edge to left and right nodes
+
+        Parameters
+        ----------
+        node_left : GremlinScenarioNode
+            One side of the edge
+        node_right : GremlinScenarioNode
+            One side of the edge
+
+        Raises
+        ------
+        GremlinParameterError
+            If the node_left or node_right are not of the type GremlinScenarioNode
+        """
+        self._validate_type(node_left)
+        self._validate_type(node_right)
+        node_left._edges.pop(node_right.id)
+        node_right._edges.pop(node_left.id)
 
     def append(self, new_node: GremlinScenarioNode) -> None:
+        """
+        Adds new_node to the graph.
+
+        Duplicates the functionality of list.append() and ensures the `head` node is set.
+
+        Parameters
+        ----------
+        new_node : GremlinScenarioNode
+            Node to add to the graph
+
+        Raises
+        ------
+        GremlinParameterError
+            If the new_node are not of the type GremlinScenarioNode
+        """
         self._validate_type(new_node)
         if self.head is None:
-            new_node.next = new_node
-            new_node.previous = new_node
             self.head = new_node
-        else:
-            self.insert_after(self.head.previous, new_node)
+        self._nodes.append(new_node)
 
-    def get_node(self, _index: int) -> Optional[GremlinScenarioNode]:
-        if not isinstance(_index, int):
-            error_msg: str = (
-                f"get_node expects index as integer, received {type(_index)}"
+    def get_node(self, uid: str) -> Optional[GremlinScenarioNode]:
+        """
+        Retrieves the GremlinScenarioNode, if it exists, with id `id`
+
+        Parameters
+        ----------
+        uid : str
+            Node id to retrieve from graph
+        """
+        for node in self._nodes:
+            if node.id == uid:
+                return node
+        return None
+
+    def get_nodes_linear(
+        self,
+        node: GremlinScenarioNode = None,
+        parent_id: str = None,
+        next_index: int = 0,
+    ) -> dict:
+        """
+        Retrieves all nodes, ordered by their defined edges.
+
+        Only will work for a linear-defined graph (node -> node_2 -> node_3)
+
+        Parameters
+        ----------
+        node : GremlinScenarioNode optional
+            Optional root node from which to begin collecting ordered node list.
+            Defaults to the `head` node if none is supplied.
+        parent_id : str optional
+            The trailing node id to be skipped when enumerating a node's edges.
+            Prevents recursion loops
+        next_index : int optional
+            The next index to use
+        """
+        if not node:
+            return self.get_nodes_linear(self.head)
+        self._validate_type(node)
+        node.index = str(next_index)
+        node.next = str(next_index + 1)
+        nodes: dict = {str(next_index): node.data}
+        print(node.id)
+        for node_id in node._edges:
+            if node_id == parent_id:
+                continue
+            nodes.update(
+                self.get_nodes_linear(
+                    node._edges[node_id]["node"], node.id, next_index + 1
+                )
             )
-            log.error(error_msg)
-            raise GremlinParameterError(error_msg)
-        node: GremlinScenarioNode = self.head  # type: ignore
-        for idx in range(_index):
-            node = node.next
-            if node == self.head:
-                return None
-        return node
+        return nodes
 
+    def get_nodes_parallel(self):
+        raise NotImplementedError("Parallel Scenario Nodes NOT IMPLEMENTED")
+
+    def insert_between(
+        self,
+        node: GremlinScenarioNode,
+        node_left: GremlinScenarioNode,
+        node_right: GremlinScenarioNode,
+    ) -> None:
+        raise NotImplementedError("insert_between NOT IMPLEMENTED")
+
+    def total_nodes(self) -> int:
+        raise NotImplementedError("total_nodes NOT IMPLEMENTED")
+
+    def longest_path(self) -> Tuple[str, str, int]:
+        raise NotImplementedError("longest_path NOT IMPLEMENTED")
+
+    @deprecated("Use add_edge instead")
     def insert_after(
         self, ref_node: GremlinScenarioNode, new_node: GremlinScenarioNode
     ) -> None:
         self._validate_type(new_node)
-        new_node.previous = ref_node
-        new_node.next = ref_node.next
-        new_node.next.previous = new_node
-        ref_node.next = new_node
+        self.add_edge(ref_node, new_node)
 
+    @deprecated("Use add_edge instead")
     def insert_before(
         self, ref_node: GremlinScenarioNode, new_node: GremlinScenarioNode
     ) -> None:
         self._validate_type(new_node)
-        self.insert_after(ref_node.previous, new_node)
+        self.add_edge(ref_node, new_node)
 
-    def next(self) -> None:
-        self.head = self.head.next
-
-    def nodes(self):
-        if self.head is None:
-            return
-        node = self.head
-        while True:
-            yield node
-            node = node.next
-            if node == self.head:
-                return
+    @deprecated
+    def nodes(self) -> list:
+        return self._nodes
 
+    @deprecated
     def nodes_data_circular(self):
-        for node in self.nodes():
+        for node in self._nodes:
             data = node.data
             yield node, data
 
+    @deprecated
     def nodes_data_linear(self):
-        for node in self.nodes():
+        for node in self._nodes:
             data = node.data
-            if node.next == self.head:
-                data.pop("next")
             yield node, data
 
+    @deprecated
     def previous(self) -> None:
-        self.head = self.head.previous
+        pass
 
     def push(self, new_node: GremlinScenarioNode) -> None:
+        """
+        Adds new_node to graph, replacing the `head` node
+
+        Parameters
+        ----------
+        new_node : GremlinScenarioNode
+            Node to add to the graph
+        """
         self._validate_type(new_node)
         self.append(new_node)
+        self.add_edge(new_node, self.head)
         self.head = new_node
 
     def remove(self, _node: GremlinScenarioNode) -> None:
-        if self.head.next == self.head:
-            self.head = None  # type: ignore
+        """
+        Removes _node from graph.
+
+        If the node to remove is the head node, Raises an error
+
+        Parameters
+        ----------
+        _node : GremlinScenarioNode
+            Node to remove from the graph
+
+        Raises
+        ------
+        GremlinParameterError
+            If the _node to remove is the current head node
+        """
+        if self.head == _node:
+            error_msg: str = f"Node to remove cannot be the current head node"
+            log.error(error_msg)
+            raise GremlinParameterError(error_msg)
         else:
-            _node.previous.next = _node.next
-            _node.next.previous = _node.previous
-            if self.head == _node:
-                self.head = _node.next
+            for node_id in _node._edges:
+                _node._edges[node_id]["node"]._edges.pop(_node.id)
+            _node._edges = {}
+            self._nodes.remove(_node)
 
     @property
     def head(self) -> GremlinScenarioNode:
